import os
import shutil
import tempfile
import unittest

import mkdocsmerge.merge

from .utils import generate_website, make_simple_yaml


class TestRunMerge(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.owd = os.getcwd()
        os.chdir(self.tmpdir)

    def test_run_merge(self):

        docs_dir_map = {
            '__master__': 'master_docs',
            'Test': 'test_docs'
        }

        site_names = ['__master__', 'Foo', 'Bar', 'Test']
        for site_name in site_names:
            docs_dir = docs_dir_map.get(site_name, None)
            yml = make_simple_yaml(site_name, docs_dir)
            generate_website(self.tmpdir, site_name, yml)

        merged_pages = mkdocsmerge.merge.run_merge(
            site_names[0], site_names[1:], True, lambda x: None)

        for site_name in site_names[1:]:
            index_path = os.path.join(
                site_name, docs_dir_map.get(site_name, 'docs'), 'index.md')
            self.assertTrue(os.path.exists(index_path))

            docs_dir_path = os.path.join(
                site_names[0], docs_dir_map.get(site_names[0], 'docs'),
                '%s_website' % site_name.lower())
            self.assertTrue(os.path.exists(docs_dir_path))

        self.assertEqual(merged_pages, {
            'site_name': '__master__ Website',
            'nav': [
                {'Home': "index.md"},
                {'Foo Website': [
                    {'Home': 'foo_website/index.md'}
                ]},
                {'Bar Website': [
                    {'Home': 'bar_website/index.md'}
                ]},
                {'Test Website': [
                    {'Home': 'test_website/index.md'}
                ]}
            ],
            'docs_dir': 'master_docs'
        })

    def tearDown(self):
        # Avoid leaving the temp directory open until program exit (bug in Windows)
        os.chdir(self.owd)
        shutil.rmtree(self.tmpdir)



import os
from ruamel.yaml import YAML


def generate_website(directory, name, yml=None):
    root = os.path.join(directory, name)
    os.mkdir(root)

    if yml is None:
        yml = make_simple_yaml(name)

    yaml = YAML()
    with open(os.path.join(root, 'mkdocs.yml'), 'w') as f:
        yaml.dump(yml, f)

    docs_folder = yml.get('docs_dir', 'docs')

    docs_dir = os.path.join(root, docs_folder)
    os.mkdir(docs_dir)
    generate_dummy_pages(docs_dir, 'nav', yml['nav'])


def make_simple_yaml(name, docs_dir=None):
    yml = {
        'site_name': '%s Website' % name,
        'nav': [
            {'Home': "index.md"},
        ]
    }

    if docs_dir:
        yml['docs_dir'] = docs_dir
    return yml


def generate_dummy_pages(docs_dir, key, node):
    if isinstance(node, list):
        for item in node:
            generate_dummy_pages(docs_dir, key, item)
    elif isinstance(node, dict):
        for k, v in node.items():
            generate_dummy_pages(docs_dir, k, v)
    else:
        path = os.path.join(docs_dir, str.replace(node, '\\', '/'))
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        with open(path, 'w') as mdf:
            mdf.write("""
* %s Title
Contents
""" % key)



import os.path
from ruamel.yaml import YAML
# Both imports are needed to avoid errors in Windows
import distutils
from distutils import dir_util

MKDOCS_YML = 'mkdocs.yml'
CONFIG_NAVIGATION = 'nav'


def run_merge(master_site, sites, unify_sites, print_func):

    # Custom argument validation instead of an ugly generic error
    if not sites:
        print_func('Please specify one or more sites to merge to the master '
                   'site.\nUse "mkdocs-merge run -h" for more information.')
        return

    # Read the mkdocs.yml from the master site
    master_yaml = os.path.join(master_site, MKDOCS_YML)
    if not os.path.isfile(master_yaml):
        print_func('Could not find the master site yml file, '
                   'make sure it exists: ' + master_yaml)
        return None

    # Round-trip yaml loader to preserve formatting and comments
    yaml = YAML()
    with open(master_yaml) as master_file:
        master_data = yaml.load(master_file)

    master_docs_dir = master_data.get('docs_dir', 'docs')
    master_docs_root = os.path.join(master_site, master_docs_dir)

    # Get all site's navigation pages and copy their files
    new_navs = merge_sites(sites, master_docs_root, unify_sites, print_func)

    # then add them to the master nav section
    master_data[CONFIG_NAVIGATION] += new_navs

    # Rewrite the master's mkdocs.yml
    with open(master_yaml, 'w') as master_file:
        yaml.dump(master_data, master_file)

    return master_data


def merge_sites(sites, master_docs_root, unify_sites, print_func):
    """
    Copies the sites content to the master_docs_root and returns
    the new merged "nav" pages to be added to the master yaml.
    """

    # NOTE: need to do this otherwise subsequent distutil.copy_tree will fail if
    # mkdocs-merge is used as a module (https://stackoverflow.com/a/28055993/920464)
    distutils.dir_util._path_created = {}

    new_navs = []
    for site in sites:
        print_func('\nAttempting to merge site: ' + site)
        site_yaml = os.path.join(site, MKDOCS_YML)
        if not os.path.isfile(site_yaml):
            print_func('Could not find the site yaml file, this site will be '
                       'skipped: "' + site_yaml + '"')
            continue

        with open(site_yaml) as site_file:
            try:
                yaml = YAML(typ='safe')
                site_data = yaml.load(site_file)
            except Exception:
                print_func('Error loading the yaml file "' + site_yaml + '". '
                           'This site will be skipped.')
                continue

        # Check 'site_data' has the 'nav' mapping
        if CONFIG_NAVIGATION not in site_data:
            print_func('Could not find the "nav" entry in the yaml file: "' +
                       site_yaml + '", this site will be skipped.')
            if 'pages' in site_data:
                raise ValueError('The site ' + site_yaml + ' has the "pages" setting in the YAML file which is not '
                                 'supported since MkDocs 1.0 and is not supported anymore by MkDocs Merge. Please '
                                 'update your site to MkDocs 1.0 or higher.')

        try:
            site_name = str(site_data['site_name'])
        except Exception:
            site_name = os.path.basename(site)
            print_func('Could not find the "site_name" property in the yaml file. '
                       'Defaulting the site folder name to: "' + site_name + '"')

        site_root = site_name.replace(' ', '_').lower()
        site_docs_dir = site_data.get('docs_dir', 'docs')

        # Copy site's files into the master site's "docs" directory
        old_site_docs = os.path.join(site, site_docs_dir)
        new_site_docs = os.path.join(master_docs_root, site_root)

        if not os.path.isdir(old_site_docs):
            print_func('Could not find the site "docs_dir" folder. This site will '
                       'be skipped: ' + old_site_docs)
            continue

        try:
            # Update if the directory already exists to allow site unification
            dir_util.copy_tree(old_site_docs, new_site_docs, update=1)
        except OSError as exc:
            print_func('Error copying files of site "' +
                       site_name + '". This site will be skipped.')
            print_func(exc.strerror)
            continue

        # Update the nav data with the new path after files have been copied
        update_navs(site_data[CONFIG_NAVIGATION], site_root, print_func=print_func)
        merge_single_site(new_navs, site_name,
                          site_data[CONFIG_NAVIGATION], unify_sites)

        # Inform the user
        print_func('Successfully merged site located in "' + site +
                   '" as sub-site "' + site_name + '"\n')

    return new_navs


def merge_single_site(global_nav, site_name, site_nav, unify_sites):
    """
    Merges a single site's nav to the global nav's data. Supports unification
    of sub-sites with the same site_name.
    """
    unified = False
    if unify_sites:
        # Check if the site_name already exists in the global_nav
        for page in global_nav:
            if site_name in page:
                # Combine the new site's pages to the existing entry
                page[site_name] = page[site_name] + site_nav
                unified = True
                break
    # Append to the global list if no unification was requested or it didn't exist.
    if (not unify_sites) or (not unified):
        global_nav.append({site_name: site_nav})


def update_navs(navs, site_root, print_func):
    """
    Recursively traverses the lists of navs (dictionaries) to update the path
    of the navs with the site_name, used as a subsection in the merged site.
    """
    if isinstance(navs, list):
        for page in navs:
            if isinstance(page, str):
                navs[navs.index(page)] = site_root + "/" + page
            else:
                update_navs(page, site_root, print_func)
    elif isinstance(navs, dict):
        for name, path in navs.items():
            if isinstance(path, str):
                navs[name] = site_root + '/' + path
            elif isinstance(path, list):
                update_navs(navs[name], site_root, print_func)
            else:
                print_func(
                    'Error merging the "nav" entry in the site: ' + site_root)
    else:
        print_func('Error merging the "nav" entry in the site: ' + site_root)



"""
Tests for the MkDocs Merge package
"""

import unittest
import mkdocsmerge.merge


class TestSiteMerges(unittest.TestCase):
    """
    Test class for MkDocs Merge. Will be separated when necessary.
    """

    def setUp(self):
        print('Test: ' + self._testMethodName)

    def test_update_pages(self):
        """
        Verifies the function that updates the path to the pages adding the another
        new subroute at the begining of each page's path
        """
        # Create original and expected data structures
        subpage = 'new_root'
        subpage_path = subpage + '/'
        nav = [{'Home': 'index.md'},
               {'About': 'menu/about.md'},
               {'Projects': [
                   {'First': 'projects/first.md'},
                   {'Nested': [
                       {'Third': 'projects/nest/third.md'}
                   ]},
                   {'Second': 'projects/second.md'}
               ]}]

        expected = [{'Home': subpage_path + 'index.md'},
                    {'About': subpage_path + 'menu/about.md'},
                    {'Projects': [
                        {'First': subpage_path + 'projects/first.md'},
                        {'Nested': [
                            {'Third': subpage_path + 'projects/nest/third.md'}
                        ]},
                        {'Second': subpage_path + 'projects/second.md'}
                    ]}]

        mkdocsmerge.merge.update_navs(nav, subpage, lambda x: None)
        self.assertEqual(nav, expected)

    def test_singe_site_merge(self):
        """
        Verifies merging of a single site's nav to the global nav's data without unification.
        """
        site_name = 'Projects'
        # Create original and expected data structures
        global_nav = [{'Home': 'index.md'},
                      {'About': 'menu/about.md'},
                      {site_name: [
                          {'First': 'projects/first.md'},
                          {'Second': 'projects/second.md'}
                      ]}]

        site_nav = [{'Nested': [
            {'Third': 'projects/nest/third.md'},
            {'Fourth': 'projects/nest/fourth.md'}
        ]}]

        expected = [{'Home': 'index.md'},
                    {'About': 'menu/about.md'},
                    {site_name: [
                        {'First': 'projects/first.md'},
                        {'Second': 'projects/second.md'},
                    ]},
                    {site_name: [
                        {'Nested': [
                            {'Third': 'projects/nest/third.md'},
                            {'Fourth': 'projects/nest/fourth.md'}
                        ]},
                    ]}]

        mkdocsmerge.merge.merge_single_site(
            global_nav, site_name, site_nav, False)
        self.assertEqual(global_nav, expected)

    def test_singe_site_merge_unified(self):
        """
        Verifies merging of a single site's nav to the global nav's data with unification
        of the sub-sites with the same site_name
        """
        site_name = 'Projects'
        # Create original and expected data structures
        global_nav = [{'Home': 'index.md'},
                      {'About': 'menu/about.md'},
                      {site_name: [
                          {'First': 'projects/first.md'},
                          {'Second': 'projects/second.md'}
                      ]}]

        site_nav = [{'Nested': [
            {'Third': 'projects/nest/third.md'},
            {'Fourth': 'projects/nest/fourth.md'}
        ]}]

        expected = [{'Home': 'index.md'},
                    {'About': 'menu/about.md'},
                    {site_name: [
                        {'First': 'projects/first.md'},
                        {'Second': 'projects/second.md'},
                        {'Nested': [
                            {'Third': 'projects/nest/third.md'},
                            {'Fourth': 'projects/nest/fourth.md'}
                        ]},
                    ]}]

        mkdocsmerge.merge.merge_single_site(
            global_nav, site_name, site_nav, True)
        self.assertEqual(global_nav, expected)

    def test_update_pages_with_section_indexes(self):
        """
        Verifies the correct updating of the section index pages' paths.
        """
        # Create original and expected data structures
        subpage = 'new_root'
        subpage_path = subpage + '/'
        nav = [{'Home': 'index.md'},
               {'Projects': [
                   'projects/index.md',
                   {'Nested': [
                       'projects/nested/index.md',
                       {'Third': 'projects/nest/third.md'}
                   ]}
               ]}]

        expected = [{'Home': subpage_path + 'index.md'},
                    {'Projects': [
                        subpage_path + 'projects/index.md',
                        {'Nested': [
                            subpage_path + 'projects/nested/index.md',
                            {'Third': subpage_path + 'projects/nest/third.md'}
                        ]}
                    ]}]

        mkdocsmerge.merge.update_navs(nav, subpage, lambda x: None)
        self.assertEqual(nav, expected)
